// ---------------------------------------------------------------------------
//
//    File created programmatically
//
//    The definitions are in the "qdarkstyle.palette" module
//
//    WARNING! All changes made in this file will be lost!
//
//----------------------------------------------------------------------------
$COLOR_BACKGROUND_LIGHT: #606060;
$COLOR_BACKGROUND_NORMAL: #535353;
$COLOR_BACKGROUND_DARK: #434343;
$COLOR_FOREGROUND_LIGHT: #F3F3F3;
$COLOR_FOREGROUND_NORMAL: #DDDDDD;
$COLOR_FOREGROUND_DARK: #CCCCCC;
$COLOR_SELECTION_LIGHT: #148CD2;
$COLOR_SELECTION_NORMAL: #1464A0;
$COLOR_SELECTION_DARK: #14506E;
$OPACITY_TOOLTIP: 230;
$SIZE_BORDER_RADIUS: 4px;
$BORDER_LIGHT: 1px solid $COLOR_BACKGROUND_LIGHT;
$BORDER_NORMAL: 1px solid $COLOR_BACKGROUND_NORMAL;
$BORDER_DARK: 1px solid $COLOR_BACKGROUND_DARK;
$BORDER_SELECTION_LIGHT: 1px solid $COLOR_SELECTION_LIGHT;
$BORDER_SELECTION_NORMAL: 1px solid $COLOR_SELECTION_NORMAL;
$BORDER_SELECTION_DARK: 1px solid $COLOR_SELECTION_DARK;
$W_STATUS_BAR_BACKGROUND_COLOR: #14506E;
$PATH_RESOURCES: '/Users/zlatko.minevibm.com/Box Sync/ZLATKO_IBM/ZLATKO_IBM/Repo/CODE/IBMQ_repos/qiskit_metal/qiskit_metal/_gui/styles/metal3';
