# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
=================================================
QGeometryTables (:mod:`qiskit_metal.qgeometries`)
=================================================

.. currentmodule:: qiskit_metal.qgeometries

This modules defines and handles everything related to the qgeometry abstraction.

QGeometryTables
---------------

.. autosummary::
    :toctree: .

    QGeometryTables

"""
from .qgeometries_handler import is_qgeometry_table, QGeometryTables  # , QGeometry Types
