import numpy as np

MHzRad = 2 * np.pi * 1e6
GHzRad = 2 * np.pi * 1e9

NANO = 1e-9
FEMTO = 1e-15
ONE_OVER_FEMTO = 1e15

e = 1.60217657e-19  # electron charge
h = 6.62606957e-34  # Plank's
hbar = 1.0545718E-34  # Plank's reduced
phinot = 2.067 * 1E-15  # magnetic flux quantum
phi0 = phinot / (2 * np.pi)  # reduced magnetic flux quantum